// StarPong v 0.95b (c) 1998 Graham Mitchell <mitchell@onr.com> - 22 Nov 1998

// main.cpp
// main program

/*
	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License (LICENSE.TXT) for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software Foundation,
	Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "reduce.h"   // speeds compile time a hair
#include <stdio.h>    // need printf and sprintf
#include <stdlib.h>   // need rand and srand
#include <time.h>     // need time() for srand
#include <allegro.h>
#include "const.h"    // global constants
#include "trigtab.h"  // sin and cos tables
#include "palette.h"  // palette generator and color constants
#include "sound.h"    // simple sound functions blip() and chirp()
#include "ball.h"     // Ball class
#include "paddle.h"   // Paddle class


// prototypes
void dostars(BITMAP*, int=1, int=WHITE);
void init();
void process_flags(int, char**);
void intro();
void flash(int,int);
void check_toggles();
void display_help();
void update_offset(int);
void declare_winner(int,int);
void outro();
void exit();


// global variables
volatile unsigned long int target_cycles = 0;
unsigned long int actual_cycles = 0;
bool countscore = true;
bool usevsync = true;
bool movingbkgd = true;
bool paused = false;
int f2offset = 0;
int f3offset = 0;
BITMAP *buffer;
BITMAP *field;
BITMAP *field2;
BITMAP *field3;
BITMAP *fieldall;
PALETTE pal;


int main(int argc, char** argv)
{
	srand(unsigned(time(NULL)));
	if ( argc>1 ) process_flags(argc,argv);
	init();

	bool scored = false;
	bool quit = false;
	int collision_flag;
	int ball_delay = 0;
	int dx, dy;
	Ball b;
	Paddle p, q;

	const int len=30;
	q.set_loc(6,(SCREEN_H-len)/2);
	q.set_keys(KEY_W,KEY_S,KEY_A,KEY_D);
	q.set_score_loc(9,2);
	q.set_goal_loc(1,SCREEN_H/2-len*3,5,SCREEN_H/2+len*3);

	actual_cycles = target_cycles = 0;

	// begin main game loop
	while ( !quit )
	{
		// generate one frame
		if ( !paused )
		{
			// generate the moving background...
			if ( movingbkgd )
			{
				blit(field, buffer, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
				masked_blit(field2,buffer,f2offset,0,0,0,SCREEN_W,SCREEN_H);
				masked_blit(field3,buffer,f3offset,0,0,0,SCREEN_W,SCREEN_H);
				update_offset( BALL_LEFT );
			}
			// or the static one.
			else
				blit(fieldall, buffer, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

			// generate all the paddles and the ball
			p.draw(buffer);
			q.draw(buffer);
			b.draw(buffer);
		}
		// and draw the whole screen in one fell swoop
		if ( usevsync ) vsync();
		blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

		while ( target_cycles > actual_cycles )
		{
			// now perform one cycle of game logic

			if ( !paused )
			{
				// move everybody
				p.move();
				q.move();
				if ( ! ball_delay )
					b.move();

				// check for collisions with paddles
				dx = dy = 0;
				collision_flag = b.touches ( p.location() ) ;
				if ( collision_flag )
				{
					p.get_momentum(dx,dy);
					b.receive_momentum(double(dx),double(dy),collision_flag);
				}
				collision_flag = b.touches ( q.location() ) ;
				if ( collision_flag )
				{
					q.get_momentum(dx,dy);
					b.receive_momentum(double(dx),double(dy),collision_flag);
				}
				// and with the scoring box
				if ( countscore )
				{
					if ( b.touches ( p.goal_loc() ) )
						{ q.point(); scored = true; }
					if ( b.touches ( q.goal_loc() ) )
						{ p.point(); scored = true; }
					if ( scored )
					{
						p.gohome();
						q.gohome();
						flash ( int(b.location().x), int(b.location().y) );
						b.score();
						buzzer();
						flash ( int(b.location().x), int(b.location().y) );
						ball_delay = GAME_SEC/2;
						scored = false;
					}
				}
			} // end pause

			check_toggles();
			++actual_cycles;
			if ( ball_delay ) --ball_delay;
			// end one game cycle
		}

		quit = ( key[KEY_ESC] || p.points() >= 10 || q.points() >= 10 );
	} // end main game loop

	if ( countscore )
		declare_winner(p.points(), q.points());
	exit();

	printf("\nStarPong v 0.95b (c) 1998 Graham Mitchell - 22 Nov 1998\nhttp://www.onr.com/user/mitchell/\n");

	return 0;
}


// timer interrupt handler
void pacer(...)
{
	target_cycles++;
}
END_OF_FUNCTION(pacer)


#include <string.h>

// handle arguments from the command line
void process_flags(int argc, char* argv[])
{
	for ( int i=1; i<argc; ++i )
	{
		if ( strnicmp(&argv[i][1],"score=",6) == 0 )
			countscore = ( stricmp(&argv[i][7],"on") == 0 );
		else if ( strnicmp(&argv[i][1],"vsync=",6) == 0 )
			usevsync = ( stricmp(&argv[i][7],"on") == 0 );
		else if ( strnicmp(&argv[i][1],"sound=",6) == 0 )
			usesound = ( stricmp(&argv[i][7],"on") == 0 );
		else if ( strnicmp(&argv[i][1],"motion=",7) == 0 )
			movingbkgd = ( stricmp(&argv[i][7],"on") == 0 );
	}

}


// initialization
void init()
{

	allegro_init();
	install_keyboard();

	// initialize sin and cos tables
	trig_table_init();

	// timing code setup
	install_timer();
	LOCK_FUNCTION(pacer);
	LOCK_VARIABLE(target_cycles);
	install_int_ex(pacer, BPS_TO_TIMER(TARGET_FPS));

	// color and palette
	set_color_depth(8);
	set_gfx_mode(GFX_VGA, 320, 200, 0, 0);
	mkpal(pal);
	set_palette(pal);

	// create static field of darkest stars in background
	field = create_bitmap(SCREEN_W, SCREEN_H);
	dostars(field,1,GREY08);
	rect(field,0,0,SCREEN_W-1,SCREEN_H-1,WHITE);

	// create slow-moving field of dim stars in middle layer
	field2 = create_bitmap(SCREEN_W*3, SCREEN_H);
	dostars(field2,3,GREY16);
	blit(field2, field2, 0, 0, SCREEN_W*2, 0, SCREEN_W, SCREEN_H);
	f2offset=SCREEN_W;

	// create fast-moving field of bright stars in front layer
	field3 = create_bitmap(SCREEN_W*3, SCREEN_H);
	dostars(field3,3);
	blit(field3, field3, 0, 0, SCREEN_W*2, 0, SCREEN_W, SCREEN_H);
	f3offset=SCREEN_W;

	// create single static bitmap with all three layers combined
	//    (this saves a lot of blitting when motion is turned off)
	fieldall = create_bitmap(SCREEN_W, SCREEN_H);
	blit(field, fieldall, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
	masked_blit(field2, fieldall, f2offset, 0, 0, 0, SCREEN_W, SCREEN_H);
	masked_blit(field3, fieldall, f3offset, 0, 0, 0, SCREEN_W, SCREEN_H);

	// create buffer to double-buffer into
	buffer = create_bitmap(SCREEN_W, SCREEN_H);
	clear_to_color(buffer, BLANK);

	// and clear the screen just for good measure
	clear_to_color(screen, BLANK);

}


// toggle state of vsync, scoring, pause, etc.
void check_toggles()
{
	if ( key[KEY_PAUSE] )
	{
		// pause the game
		clear_keybuf();
		textout_centre(buffer,font,"PAUSED",SCREEN_W/2,SCREEN_H/2-5,GREY16);
		paused = !paused;
	}
	if ( key[KEY_F1] )
	{
		// show help screen
		clear_keybuf();
		display_help();
		while ( !readkey() )
			;
		if ( key[KEY_F1] )
			clear_keybuf();
		actual_cycles = target_cycles;
	}
	if ( key[KEY_F9] )
	{
		// toggle sound
		clear_keybuf();
		usesound = !usesound;
	}
	if ( key[KEY_F10] )
	{
		// toggle scoring
		clear_keybuf();
		countscore = !countscore;
	}
	if ( key[KEY_F11] )
	{
		// toggle vsync
		clear_keybuf();
		usevsync = !usevsync;
	}
	if ( key[KEY_F12] )
	{
		// toggle moving background
		clear_keybuf();
		movingbkgd = !movingbkgd;
		// save current background in 'fieldall' to cut down on useless blits
		if ( !movingbkgd )
		{
			blit(field, fieldall, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
			masked_blit(field2,fieldall,f2offset,0,0,0,SCREEN_W,SCREEN_H);
			masked_blit(field3,fieldall,f3offset,0,0,0,SCREEN_W,SCREEN_H);
		}
	}
}


// show basic help screen
void display_help()
{
	const int xbord = 30;
	const int ybord = 35;
	const int x = xbord+20;
	int y = ybord+8;
	const int d = 11;
	const int c = WHITE;
	const int lines = 7;

	rectfill(screen, xbord, ybord, SCREEN_W-xbord, y+(lines*d)+16, GREY16);
	rect(screen, xbord, ybord, SCREEN_W-xbord, y+(lines*d)+16, WHITE);
	text_mode(-1);
	textout(screen, font, "StarPong v 0.95 beta", xbord+4, ybord+4, c); 
	textout(screen, font, "F1    - this help menu", x, y+=d, c); 
	textout(screen, font, "F9    - toggle sound on/off", x, y+=d, c); 
	textout(screen, font, "F10   - toggle scoring on/off", x, y+=d, c); 
	textout(screen, font, "F11   - toggle vsync on/off", x, y+=d, c); 
	textout(screen, font, "F12   - moving stars on/off", x, y+=d, c); 
	textout(screen, font, "PAUSE - pause the game", x, y+=d, c); 
	textout(screen, font, "ESC   - quit", x, y+=d, c); 
}


// make a quick screen flash
void flash( int x, int y )
{
	int r = 2;
	for ( int c=GREY32; c>GREY00; --c)
		circle(screen, x, y, ++r, c);
	rest(10);
}


// move front and middle star field bitmaps depending on ball's direction
void update_offset( int dir )
{
	if ( dir == BALL_LEFT )
	{
		++f3offset;
		if ( f3offset%2 == 0 ) ++f2offset;
		if ( f2offset >= SCREEN_W*2 )
			f2offset = 0;
		if ( f3offset >= SCREEN_W*2 )
			f3offset = 0;
	}
	else
	{
		--f3offset;
		if ( f3offset%2 == 0 ) --f2offset;
		if ( f2offset <= 0 )
			f2offset = SCREEN_W*2;
		if ( f3offset <= 0 )
			f3offset = SCREEN_W*2;
	}
}


// display who won
void declare_winner(int a, int b)
{
	char temp[10];
	char* p1win = "Right player is the winner!!";
	char* p2win = "Left player is the winner!!";
	char* tiemsg = "It's a tie!!";
	char* msg;

	blit(field3, buffer, f3offset, 0, 0, 0, SCREEN_W, SCREEN_H);
	masked_blit(field2, buffer, f2offset, 0, 0, 0, SCREEN_W, SCREEN_H);
	masked_blit(field, buffer, 0, 0, 0, 0, SCREEN_W, SCREEN_H);

	if ( a > b ) msg = p1win;
	else if ( a < b ) msg = p2win;
	else msg = tiemsg;
		
	textout_centre(buffer,font,msg,SCREEN_W/2,SCREEN_H/2-5,WHITE);
	sprintf(temp, "%d   %d",b,a);
	textout_centre(buffer,font,temp,SCREEN_W/2,SCREEN_H/2+5,WHITE);
	blit(buffer, screen, 0, 0, 0, 0, SCREEN_W, SCREEN_H);
	rest(3000);
}


// clean up and go home
void exit()
{
	destroy_bitmap(buffer);
	destroy_bitmap(field);
	destroy_bitmap(field2);
	destroy_bitmap(field3);
	destroy_bitmap(fieldall);
	allegro_exit();
}


// make star field background
void dostars( BITMAP* bmp, int mult, int color )
{
	int i,x,y;

	clear_to_color(bmp, BLANK);
	for ( i=0 ; i<STAR_DENSITY*mult ; ++i )
	{
		x = rand()%SCREEN_W*mult;
		y = rand()%SCREEN_H;
		putpixel(bmp, x, y, color);
	}

}

